/*********************************************************************
*                SEGGER Microcontroller GmbH & Co. KG                *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2019 SEGGER Microcontroller GmbH & Co. KG         *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.38                                             *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Configuration
*/

#ifndef   ALLOW_NESTED_INTERRUPTS       // Modify this value according to your needs. May be overwritten by project settings.
  #define ALLOW_NESTED_INTERRUPTS  (0)  // CAUTION: Enabling nested interrupts drastically increases stack usage!
#endif

#ifndef   USE_ROM_VECTOR_TABLE          // Modification to this value also requires modification to ventry.s! Consider using project options instead.
  #define USE_ROM_VECTOR_TABLE     (1)  // Configures ISR handling for vectored mode, in which a ROM vector table will be used for local interrupts.
#endif

/*********************************************************************
*
*       System tick settings
*/

#define OS_TIMER_FREQ    (32768uL)
#define OS_TICK_FREQ     ( 1000uL)
#define OS_INT_FREQ      ( 1024uL)

#define OS_TIMER_RELOAD  (OS_TIMER_FREQ / OS_INT_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifdef OS_LIBMODE_SAFE
  #undef    OS_VIEW_IFSELECT
  #define   OS_VIEW_IFSELECT  OS_VIEW_DISABLED  // Communication not supported in OS_LIBMODE_SAFE
#else
  #ifndef   OS_VIEW_IFSELECT
    #define OS_VIEW_IFSELECT  OS_VIEW_DISABLED
  #endif
#endif

/*********************************************************************
*
*       Local defines (sfrs and addresses used in RTOSInit.c)
*
**********************************************************************
*/

//
//  Coreplex-Local Interrupts
//
#define COREPLEX_BASE_ADDR             (0x02000000u)
#define COREPLEX_MSIP                  (*(volatile OS_U32*)(COREPLEX_BASE_ADDR + 0x0000u + (8u * _OS_RISCV_HartID)))  // LSB of this register may be used to generate SW interrupts
#define COREPLEX_MTIMECMP              (*(volatile OS_U64*)(COREPLEX_BASE_ADDR + 0x4000u + (8u * _OS_RISCV_HartID)))  // Used to generate OS tick, Timer compare register
#define COREPLEX_MTIME                 (*(volatile OS_U64*)(COREPLEX_BASE_ADDR + 0xBFF8u))                            // Used to generate OS tick, Timer counter register

#define COREPLEX_LOCAL_NUM_INTERRUPTS  (32u)

//
// Coreplex Platform-Level Interrupt Controller
//
#define COREPLEX_PLIC_BASE_ADDR        (0x0C000000u)

#define COREPLEX_PLIC_NUM_INTERRUPTS   (52u)
#define COREPLEX_PLIC_NUM_INT_PRIOS    ( 7u)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_U32 _OS_RISCV_HartID = 0u;

#if (USE_ROM_VECTOR_TABLE == 0)
static OS_ISR_HANDLER* _apLocalIsrHandler[COREPLEX_LOCAL_NUM_INTERRUPTS]   __attribute__ ((aligned (4), section (".non_init")));
#endif
static OS_ISR_HANDLER* _apExternalIsrHandler[COREPLEX_PLIC_NUM_INTERRUPTS] __attribute__ ((aligned (4), section (".non_init")));

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  OS_I64 Cnt;

  Cnt = COREPLEX_MTIMECMP - COREPLEX_MTIME;
  if (Cnt > 0) {
    return (unsigned int)(OS_TIMER_RELOAD - Cnt);
  } else {
    return (unsigned int)0 - Cnt;
  }
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  if (COREPLEX_MTIMECMP - COREPLEX_MTIME > 0u) {
    return (unsigned int)0u;
  } else {
    return (unsigned int)1u;
  }
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_ISR_Undefined()
*
*  Function description
*    This routine may be used to detect non-installed interrupts.
*
*  Additional information
*    OS_ISR_Undefined() is called when an interrupt is pending for
*    which no specific interrupt handler was previously installed.
*
*    In direct mode, this routine is called by OS_IRQ_Handler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void OS_ISR_Undefined(void) {
  volatile int Dummy;

  Dummy = 1;
  while (Dummy > 0) {
    //
    // You may set a breakpoint here to detect non-installed ISR-handler
    //
  }
}

/*********************************************************************
*
*       OS_ISR_Timer()
*
*  Function description
*    This routine performs the embOS tick handling.
*
*  Additional information
*    OS_ISR_Timer() is called when the Machine Timer interrupt is pending.
*    Machine Timer Interrupt becomes pending when (COREPLEX_MTIMECMP >= COREPLEX_MTIME).
*
*    In direct mode, this routine is called by OS_IRQ_Handler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void OS_ISR_Timer(void) {
  OS_U64 Compare;

  Compare = COREPLEX_MTIMECMP;                   // Read time compare value

#if (ALLOW_NESTED_INTERRUPTS != 0)
  OS_RISCV_ISR_Disable(IRQ_M_TIMER);             // We must inhibit further occurrences of this specific IRQ, for we will not clear its pending bit until later
  OS_EnterNestableInterrupt();                   // Now interrupts may be re-enabled. If nesting should be allowed
#else
  OS_EnterInterrupt();                           // Inform OS that interrupt handler is running
#endif

  do {                                           // We might have to perform numerous ticks when interrupts have been disabled for extended periods
    OS_TICK_HandleEx();
    Compare    += OS_TIMER_RELOAD;
  } while (Compare <= COREPLEX_MTIME);
  COREPLEX_MTIMECMP = Compare;                   // Eventually, configure new compare value. Clears pending bit as well.

#if (ALLOW_NESTED_INTERRUPTS != 0)
  OS_RISCV_ISR_Enable(IRQ_M_TIMER);              // Re-allow occurrences of this IRQ
  OS_LeaveNestableInterrupt();                   // Leave nestable interrupt, perform task switch if required
#else
  OS_LeaveInterrupt();                           // Leave interrupt, perform task switch if required
#endif
}

/*********************************************************************
*
*       OS_ISR_External()
*
*  Function description
*    This routine detects the specific reason for the external IRQ and
*    calls the respective interrupt service routine that was previously
*    installed by OS_RISCV_COREPLEX_ISR_InstallHandler().
*
*  Additional information
*    OS_ISR_External() is called when the Machine External interrupt is
*    pending. Machine External interrupt becomes pending when it is issued
*    by the external Platform-Level interrupt controller (PLIC).
*
*    In direct mode, this routine is called by OS_IRQ_Handler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void OS_ISR_External(void) {
  OS_U32 ISRIndex;

  ISRIndex = OS_RISCV_COREPLEX_ISR_Claim();         // Claim external interrupt by reading its ID, also clears pending bit

  if (ISRIndex != 0u) {                             // "0" indicates no IRQ, e.g. when it was already claimed by another hart
#if (ALLOW_NESTED_INTERRUPTS != 0)
    OS_U32 Threshold;
    OS_U32 Priority;

    //
    // Current threshold is always smaller than the priority of this ISR, otherwise we wouldn't be here.
    // Hence, we can configure this ISR's priority as new priority threshold while interrupts are still disabled globally.
    //
    Threshold = OS_RISCV_COREPLEX_ISR_GetThreshold();
    Priority  = OS_RISCV_COREPLEX_ISR_GetPriority(ISRIndex);
    OS_RISCV_COREPLEX_ISR_SetThreshold(Priority);   // Configure new priority threshold
    OS_EnterNestableInterrupt();                    // Now interrupts may be re-enabled. If nesting should be allowed
#else
    OS_EnterInterrupt();                            // Inform OS that interrupt handler is running
#endif

    _apExternalIsrHandler[ISRIndex]();              // Call appropriate handler
    OS_RISCV_COREPLEX_ISR_Complete(ISRIndex);       // Signal ISR completion

#if (ALLOW_NESTED_INTERRUPTS != 0)
    OS_RISCV_COREPLEX_ISR_SetThreshold(Threshold);  // Configure priority threshold back to its original value
    OS_LeaveNestableInterrupt();                    // Leave nestable interrupt, perform task switch if required
#else
    OS_LeaveInterrupt();                            // Leave interrupt, perform task switch if required
#endif
  }
}

#if (USE_ROM_VECTOR_TABLE == 0)
/*********************************************************************
*
*       OS_IRQ_Handler()
*
*  Function description
*    This routine detects the specific reason for the local IRQ and
*    calls the respective interrupt service routine that was previously
*    installed by OS_RISCV_ISR_InstallHandler().
*
*  Additional information
*    In direct mode, this routine is called from the trap_entry() function
*    found in ventry.s, and is used for all interrupt requests of any kind.
*    In vectored mode, this routine is not used, since local interrupts are
*    delegated through a ROM vector table instead.
*/
OS_U32 OS_IRQ_Handler(OS_U32 mcause, OS_U32 mepc) {
  if (mcause & MCAUSE_INT) {  // Trap caused by interrupt, call appropriate handler.
    _apLocalIsrHandler[mcause & MCAUSE_CAUSE]();
  } else {                    // Trap caused by synchronous exception, call fault handler.
    OS_ISR_SynchronousTrap(mcause, mepc);
  }
  return mepc;
}
#endif  // (USE_ROM_VECTOR_TABLE == 0)

/*********************************************************************
*
*       OS_ISR_SynchronousTrap()
*
*  Function description
*    This routine may be used to investigate synchronous traps.
*
*  Additional information
*    OS_ISR_SynchronousTrap() is called when a Synchronous Trap has occurred.
*    This is a severe error, halt execution!
*
*    In direct mode, this routine is called from OS_IRQ_Handler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void OS_ISR_SynchronousTrap(OS_U32 mcause, OS_U32 mepc) {
  OS_USEPARA(mepc);                 // May be used for further investigation.

#if (OS_DEBUG != 0)
  switch (mcause & MCAUSE_CAUSE) {  // Detect specific exception cause:
  case  0: while (1);               // - Instruction address misaligned.
  case  1: while (1);               // - Instruction access fault.
  case  2: while (1);               // - Illegal instruction.
  case  3: while (1);               // - Breakpoint.
  case  4: while (1);               // - Load address misaligned.
  case  5: while (1);               // - Load access fault.
  case  6: while (1);               // - Store/AMO address misaligned.
  case  7: while (1);               // - Store/AMO access fault.
  case  8: while (1);               // - Environment call from U-mode.
  case 11: while (1);               // - Environment call from M-mode.
  default: while (1);               // - Reserved. This should never happen.
  }
#else
  OS_USEPARA(mcause);
  while (1);  // Halt execution
#endif
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, 1, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
  OS_U32             i;

  OS_IncDI();
  //
  // Setup values for usec precise system time functions
  //
  OS_Config_SysTimer(&SysTimerConfig);
  OS_TICK_Config(OS_TICK_FREQ, OS_INT_FREQ);
  //
  // Set the ID for this hart
  //
  _OS_RISCV_HartID = OS_RISCV_GetHartID();

  //
  // ***** Initialize RISC-V interrupt handling, local *****
  //
#if (USE_ROM_VECTOR_TABLE == 0)
  //
  // Initialize direct ISR handling: install dummy handlers to detect non-installed interrupts
  //
  OS_RISCV_ISR_Init(COREPLEX_LOCAL_NUM_INTERRUPTS, _apLocalIsrHandler);  // Implicitly disables all sources
  for (i = 0u; i < COREPLEX_LOCAL_NUM_INTERRUPTS; i++) {
    OS_RISCV_ISR_InstallHandler(i, OS_ISR_Undefined);              // When in direct mode, install dummy handler to detect non-installed local interrupts
  }
#else
  //
  // Initialize vectored ISR handling: initially disable all sources (can't assume these are reset)
  //
  OS_RISCV_ISR_Disable(IRQ_M_SOFTWARE);                            // Disable Machine Software Interrupt
  OS_RISCV_ISR_Disable(IRQ_M_TIMER);                               // Disable Machine Timer Interrupt
  OS_RISCV_ISR_Disable(IRQ_M_EXTERNAL);                            // Disable Machine External Interrupt
  for (i = IRQ_LOCAL0; i < COREPLEX_LOCAL_NUM_INTERRUPTS; i++) {
    OS_RISCV_ISR_Disable(i);                                       // Disable local interrupt 0 to 15 (if available)
  }
#endif
  //
  // Configure OS timer and enable Machine Timer Interrupt
  //
  COREPLEX_MTIME    = 0;                                           // Configure counter register
  COREPLEX_MTIMECMP = OS_TIMER_RELOAD;                             // Configure compare register
#if (USE_ROM_VECTOR_TABLE == 0)
  OS_RISCV_ISR_InstallHandler(IRQ_M_TIMER, OS_ISR_Timer);          // When in direct mode, install handler for Machine Timer Interrupt
#endif
  OS_RISCV_ISR_Enable(IRQ_M_TIMER);                                // Enable Machine Timer Interrupt
  //
  // Install and enable some additional local interrupts (for demonstration on how to use local interrupt sources)
  //
#if (USE_ROM_VECTOR_TABLE == 0)
  OS_RISCV_ISR_InstallHandler(IRQ_M_EXTERNAL, OS_ISR_External);    // When in direct mode, install handler for Machine External Interrupt
#endif
  OS_RISCV_ISR_Enable(IRQ_M_EXTERNAL);                             // Enable Machine External Interrupt

  //
  // ***** Initialize RISC-V interrupt handling, external *****
  //
  //
  // Install dummy handler to detect non-installed interrupts
  //
  OS_RISCV_COREPLEX_ISR_Init(COREPLEX_PLIC_BASE_ADDR, COREPLEX_PLIC_NUM_INTERRUPTS, COREPLEX_PLIC_NUM_INT_PRIOS, _apExternalIsrHandler);  // Implicitly disables all sources
  for (i = 0u; i < COREPLEX_PLIC_NUM_INTERRUPTS; i++) {
    OS_RISCV_COREPLEX_ISR_InstallHandler(i, OS_ISR_Undefined);     // Install dummy handler to detect non-installed external interrupts
  }
  //
  // Install and enable one external interrupt (for demonstration on how to use external interrupt sources)
  //
  //
  // Configure and initialize SEGGER SystemView
  //
#if (OS_PROFILE != 0)
  SEGGER_SYSVIEW_Conf();
#endif
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {     // Idle loop: No task is ready to execute
  while (1) {            // Nothing to do ... wait for interrupt
    ;
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*  Function description
*    This routine is required for task-info via embOSView or high
*    resolution time measurement functions.
*    It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  return COREPLEX_MTIME;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*  Function description
*    Convert Cycles into micro seconds.
*
*    If your clock frequency is not a multiple of 1 MHz,
*    you may have to modify this routine in order to get proper
*    diagnostics.
*
*    This routine is required for profiling or high resolution time
*    measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  OS_U32 t;

  t = ((OS_U64)Cycles * 1000000u) / OS_TIMER_FREQ;
  return t;
}

/*************************** End of file ****************************/
